@@ -1,17 +1,25 @@
+from .exceptions import AutoCommitError
 from .exceptions import ConsumerUnhealthyException
+from .exceptions import ProducerUnhealthyException
 from .exceptions import SchemaConflictException
 from .exceptions import StopConsumer
 from .exceptions import UnhandledMessage
-from .exceptions import UnregisteredSchemaException
 from .kafka import KafkaTopicManager
 from .metrics import CONSUMED_MESSAGE_TIME
 from .metrics import CONSUMED_MESSAGES
+from .metrics import CONSUMER_REBALANCED
 from .metrics import CONSUMER_TOPIC_OFFSET
 from .metrics import MESSAGE_LEAD_TIME
 from .metrics import NOERROR
 from .metrics import PRODUCER_TOPIC_OFFSET
 from .metrics import PUBLISHED_MESSAGES
+from .metrics import PUBLISHED_MESSAGES_TIME
+from .metrics import watch_kafka
+from .metrics import watch_publish
+from .retry import RetryHandler
+from .retry import RetryPolicy
 from .utils import resolve_dotted_name
+from aiokafka.structs import TopicPartition
 from asyncio.futures import Future
 from functools import partial
 from opentracing.scope_managers.contextvars import ContextVarsScopeManager
@@ -45,10 +53,18 @@ logger = logging.getLogger("kafkaesk")
 
 
 class Subscription:
-    def __init__(self, stream_id: str, func: Callable, group: str):
+    def __init__(
+        self,
+        stream_id: str,
+        func: Callable,
+        group: str,
+        *,
+        retry_handlers: Optional[Dict[Type[Exception], RetryHandler]] = None,
+    ):
         self.stream_id = stream_id
         self.func = func
         self.group = group
+        self.retry_handlers = retry_handlers
 
     def __repr__(self) -> str:
         return f"<Subscription stream: {self.stream_id} >"
@@ -79,6 +95,8 @@ def _pydantic_msg_handler(
     try:
         return model.parse_obj(data["data"])
     except ValidationError:
+        # log the execption so we can see what fields failed
+        logger.warning(f"Error parsing pydantic model:{model} {record}", exc_info=True)
         raise UnhandledMessage(f"Error parsing data: {model}")
 
 
@@ -98,8 +116,9 @@ def _record_msg_handler(
     return record
 
 
-def _published_callback(topic: str, fut: Future) -> None:
+def _published_callback(topic: str, start_time: float, fut: Future) -> None:
     # Record the metrics
+    finish_time = time.time()
     exception = fut.exception()
     if exception:
         error = str(exception.__class__.__name__)
@@ -109,10 +128,39 @@ def _published_callback(topic: str, fut: Future) -> None:
     metadata = fut.result()
     PUBLISHED_MESSAGES.labels(stream_id=topic, partition=metadata.partition, error=error).inc()
     PRODUCER_TOPIC_OFFSET.labels(stream_id=topic, partition=metadata.partition).set(metadata.offset)
+    PUBLISHED_MESSAGES_TIME.labels(stream_id=topic).observe(finish_time - start_time)
+
+
+_aiokafka_consumer_settings = (
+    "fetch_max_wait_ms",
+    "fetch_max_bytes",
+    "fetch_min_bytes",
+    "max_partition_fetch_bytes",
+    "request_timeout_ms",
+    "auto_offset_reset",
+    "metadata_max_age_ms",
+    "max_poll_interval_ms",
+    "rebalance_timeout_ms",
+    "session_timeout_ms",
+    "heartbeat_interval_ms",
+    "consumer_timeout_ms",
+    "max_poll_records",
+    "connections_max_idle_ms",
+)
+_aiokafka_producer_settings = (
+    "metadata_max_age_ms",
+    "request_timeout_ms",
+    "max_batch_size",
+    "max_request_size",
+    "send_backoff_ms",
+    "retry_backoff_ms",
+)
 
 
 class SubscriptionConsumer:
     _consumer: Optional[aiokafka.AIOKafkaConsumer] = None
+    _to_commit: Dict[TopicPartition, int]
+    _auto_commit_task: Optional[asyncio.Task] = None
 
     def __init__(
         self,
@@ -123,11 +171,24 @@ class SubscriptionConsumer:
         self._app = app
         self._subscription = subscription
         self._event_handlers = event_handlers or {}
+        self._to_commit = {}
+        self._commit_lock = asyncio.Lock()
 
     @property
-    def consumer(self) -> Optional[aiokafka.AIOKafkaConsumer]:
+    def consumer(self) -> aiokafka.AIOKafkaConsumer:
+        if self._consumer is None:
+            raise RuntimeError("Consumer not initialized")
         return self._consumer
 
+    async def healthy(self) -> None:
+        if self._consumer is None:
+            return
+        if self._auto_commit_task is not None and self._auto_commit_task.done():
+            raise AutoCommitError(self, "No auto commit task running")
+        if not await self._consumer._client.ready(self._consumer._coordinator.coordinator_id):
+            raise ConsumerUnhealthyException(self, "Consumer is not ready")
+        return
+
     async def emit(self, name: str, *args: Any, **kwargs: Any) -> None:
         for func in self._event_handlers.get(name, []):
             try:
@@ -135,33 +196,143 @@ class SubscriptionConsumer:
             except StopConsumer:
                 raise
             except Exception:
-                logger.warning(f"Error emitting event: {name}: {func}")
+                logger.warning(f"Error emitting event: {name}: {func}", exc_info=True)
 
     async def __call__(self) -> None:
+        await self.initialize()
         try:
             while True:
-                await self.__run()
-        except aiokafka.errors.UnrecognizedBrokerVersion:
-            logger.error("Could not determine kafka version. Exiting")
-        except aiokafka.errors.KafkaConnectionError:
-            logger.warning("Connection error", exc_info=True)
+                try:
+                    await self._run()
+                except aiokafka.errors.KafkaConnectionError:
+                    logger.warning("Connection error, retrying", exc_info=True)
+                    await asyncio.sleep(0.5)
         except (RuntimeError, asyncio.CancelledError, StopConsumer):
-            logger.info("Consumer stopped, exiting")
+            logger.debug("Consumer stopped, exiting")
+        finally:
+            await self.finalize()
 
-    async def __run(self) -> None:
-        self._consumer = aiokafka.AIOKafkaConsumer(
-            bootstrap_servers=cast(List[str], self._app._kafka_servers),
-            loop=asyncio.get_event_loop(),
-            group_id=self._subscription.group,
-            api_version=self._app._kafka_api_version,
-            **self._app._kafka_settings or {},
-        )
+    async def initialize(self) -> None:
+        if self._app.auto_commit:
+            self._auto_commit_task = asyncio.create_task(self._auto_commit())
+
+    async def finalize(self) -> None:
+        if (
+            self._app.auto_commit
+            and self._auto_commit_task is not None
+            and not self._auto_commit_task.done()
+        ):
+            self._auto_commit_task.cancel()
+
+    async def _auto_commit(self) -> None:
+        default_to = self._app.kafka_settings.get("auto_commit_interval_ms", 2000) / 1000
+        assignment = None
+
+        while True:
+            to = default_to
+            try:
+                await asyncio.sleep(to)
+
+                # pulled from auto commit handler in aiokafka that
+                # still commits on error
+
+                subscription = self.consumer._subscription.subscription
+                if subscription is not None and not subscription.active:
+                    # The subscription can change few times, so we can not rely on
+                    # flags or topic lists. For example if user changes
+                    # subscription from X to Y and back to X we still need to
+                    # rejoin group.
+                    self.consumer._coordinator.request_rejoin()
+                    subscription = self.consumer._subscription.subscription
+                if subscription is None:
+                    await asyncio.wait(
+                        [self.consumer._subscription.wait_for_subscription()],
+                        return_when=asyncio.FIRST_COMPLETED,
+                    )
+                    subscription = self.consumer._subscription.subscription
+                assert subscription is not None and subscription.active
+                auto_assigned = self.consumer._subscription.partitions_auto_assigned()
+
+                try:
+                    await self.consumer._coordinator.ensure_coordinator_known()
+                    if auto_assigned and self.consumer._coordinator.need_rejoin(subscription):
+                        new_assignment = await self.consumer._coordinator.ensure_active_group(
+                            subscription, assignment
+                        )
+                        if new_assignment is None or not new_assignment.active:
+                            continue
+                        else:
+                            assignment = new_assignment
+                    else:
+                        assignment = subscription.assignment
+
+                    assert assignment is not None and assignment.active
+
+                    to = await self.commit() or default_to
+                except aiokafka.errors.KafkaError:
+                    logger.warning("Error with autocommit", exc_info=True)
+            except (RuntimeError, asyncio.CancelledError, AssertionError):
+                logger.debug("Exiting auto commit task")
+                return
+
+    async def commit(self) -> Optional[float]:
+        """
+        Commit current state
+        """
+        async with self._commit_lock:
+            # it's possible to commit on exit as well
+            # so let's make sure to use lock here
+            return await self._commit()
+        return None
+
+    async def _commit(self) -> float:
+        now = time.monotonic()
+        interval = self._app.kafka_settings.get("auto_commit_interval_ms", 2000) / 1000
+        backoff = self._app.kafka_settings.get("retry_backoff_ms", 200) / 1000
+
+        to_commit = self._to_commit
+        try:
+            self._to_commit = {}
+            assignment = self.consumer.assignment()
+            for tp in list(to_commit.keys()):
+                if tp not in assignment:
+                    # clean if we got a new assignment
+                    del to_commit[tp]
+            await self.consumer.commit(to_commit)
+            logging.debug(f"Committed offsets: {to_commit}")
+        except aiokafka.errors.CommitFailedError:
+            # try to commit again but we need to combine
+            # what could now be in the commit from when we started
+            logging.info("Failed to commit offsets, Retrying.")
+            for tp in to_commit.keys():
+                if tp in self._to_commit:
+                    self._to_commit[tp] = max(to_commit[tp], self._to_commit[tp])
+                else:
+                    self._to_commit[tp] = to_commit[tp]
+            return backoff
+        except aiokafka.errors.IllegalStateError:
+            logging.info("Attempt to commit unassigned partition")
+        except Exception:
+            logging.exception("Unhandled exception committed offsets", exc_info=True)
+        return max(0, (now + interval) - time.monotonic())
+
+    def record_commit(self, record: aiokafka.structs.ConsumerRecord) -> None:
+        self._to_commit[TopicPartition(record.topic, record.partition)] = record.offset + 1
+
+    async def _run(self) -> None:
+        self._consumer = self._app.consumer_factory(self._subscription.group)
         pattern = fnmatch.translate(self._app.topic_mng.get_topic_id(self._subscription.stream_id))
         listener = CustomConsumerRebalanceListener(
             self._consumer, self._app, self._subscription.group
         )
         self._consumer.subscribe(pattern=pattern, listener=listener)
-        await self._consumer.start()
+
+        # Initialize subscribers retry policy
+        retry_policy = RetryPolicy(app=self._app, subscription=self._subscription,)
+        await retry_policy.initialize()
+
+        with watch_kafka("consumer_start"):
+            await self._consumer.start()
 
         msg_handler = _raw_msg_handler
         sig = inspect.signature(self._subscription.func)
@@ -205,7 +376,7 @@ class SubscriptionConsumer:
                 ).observe(lead_time)
 
                 try:
-                    logger.info(f"Handling msg: {record}")
+                    logger.debug(f"Handling msg: {record}")
                     msg_data = orjson.loads(record.value)
                     it = iter(sig.parameters.items())
                     name, _ = next(it)
@@ -218,6 +389,8 @@ class SubscriptionConsumer:
                             kwargs["record"] = record
                         elif key == "app":
                             kwargs["app"] = self._app
+                        elif key == "subscriber":
+                            kwargs["subscriber"] = self
                         elif issubclass(param.annotation, opentracing.Span):
                             kwargs[key] = context.span
 
@@ -235,26 +408,33 @@ class SubscriptionConsumer:
                         partition=record.partition,
                         group_id=self._subscription.group,
                     ).inc()
-                except UnhandledMessage:
-                    # how should we handle this? Right now, fail hard
-                    logger.warning(f"Could not process msg: {record}", exc_info=True)
-                    # We increase the error counter
+                    self.record_commit(record)
+                except Exception as err:
                     CONSUMED_MESSAGES.labels(
                         stream_id=record.topic,
                         partition=record.partition,
-                        error="UnhandledMessage",
+                        error=err.__class__.__name__,
                         group_id=self._subscription.group,
                     ).inc()
-
+                    await retry_policy(record=record, exception=err)
+                    # also commit after successful message retry handling
+                    self.record_commit(record)
+                except aiokafka.errors.ConsumerStoppedError:
+                    # Consumer is closed
+                    pass
                 finally:
-                    context.span.finish()
                     context.close()
                     await self.emit("message", record=record)
         finally:
             try:
-                await self._consumer.commit()
+                await self.commit()
             except Exception:
-                logger.info("Could not commit current offsets", exc_info=True)
+                logger.info("Could not commit on shutdown", exc_info=True)
+            try:
+                # kill retry policy now
+                await retry_policy.finalize()
+            except Exception:
+                logger.info("Cound not properly stop retry policy", exc_info=True)
             try:
                 await self._consumer.stop()
             except Exception:
@@ -289,9 +469,17 @@ class Router:
 
         self._event_handlers[name].append(handler)
 
-    def subscribe(self, stream_id: str, group: str,) -> Callable:
+    def subscribe(
+        self,
+        stream_id: str,
+        group: str,
+        *,
+        retry_handlers: Optional[Dict[Type[Exception], RetryHandler]] = None,
+    ) -> Callable:
         def inner(func: Callable) -> Callable:
-            subscription = Subscription(stream_id, func, group or func.__name__)
+            subscription = Subscription(
+                stream_id, func, group or func.__name__, retry_handlers=retry_handlers
+            )
             self._subscriptions.append(subscription)
             return func
 
@@ -299,7 +487,7 @@ class Router:
 
     def schema(
         self,
-        _id: str,
+        _id: Optional[str] = None,
         *,
         version: Optional[int] = None,
         retention: Optional[int] = None,
@@ -308,9 +496,13 @@ class Router:
         version = version or 1
 
         def inner(cls: Type[BaseModel]) -> Type[BaseModel]:
-            key = f"{_id}:{version}"
+            if _id is None:
+                type_id = cls.__name__
+            else:
+                type_id = _id
+            key = f"{type_id}:{version}"
             reg = SchemaRegistration(
-                id=_id, version=version or 1, model=cls, retention=retention, streams=streams
+                id=type_id, version=version or 1, model=cls, retention=retention, streams=streams
             )
             if key in self._schemas:
                 raise SchemaConflictException(self._schemas[key], reg)
@@ -335,12 +527,13 @@ class Application(Router):
         kafka_settings: Optional[Dict[str, Any]] = None,
         replication_factor: Optional[int] = None,
         kafka_api_version: str = "auto",
+        auto_commit: bool = True,
     ):
         super().__init__()
         self._kafka_servers = kafka_servers
         self._kafka_settings = kafka_settings
         self._producer = None
-        self._intialized = False
+        self._initialized = False
         self._locks: Dict[str, asyncio.Lock] = {}
 
         self._kafka_api_version = kafka_api_version
@@ -348,6 +541,13 @@ class Application(Router):
         self._replication_factor = replication_factor
         self._topic_mng: Optional[KafkaTopicManager] = None
         self._subscription_consumers: List[SubscriptionConsumer] = []
+        self._subscription_consumers_tasks: List[asyncio.Task] = []
+
+        self.auto_commit = auto_commit
+
+    @property
+    def kafka_settings(self) -> Dict[str, Any]:
+        return self._kafka_settings or {}
 
     def mount(self, router: Router) -> None:
         self._subscriptions.extend(router.subscriptions)
@@ -356,13 +556,9 @@ class Application(Router):
 
     async def health_check(self) -> None:
         for subscription_consumer in self._subscription_consumers:
-            if (
-                subscription_consumer.consumer is not None
-                and not await subscription_consumer.consumer._client.ready(
-                    subscription_consumer.consumer._coordinator.coordinator_id
-                )
-            ):
-                raise ConsumerUnhealthyException(subscription_consumer, "Consumer is not ready")
+            await subscription_consumer.healthy()
+        if not self.producer_healthy():
+            raise ProducerUnhealthyException(self._producer)  # type: ignore
 
     async def _call_event_handlers(self, name: str) -> None:
         handlers = self._event_handlers.get(name)
@@ -378,6 +574,7 @@ class Application(Router):
                 cast(List[str], self._kafka_servers),
                 self._topic_prefix,
                 replication_factor=self._replication_factor,
+                kafka_api_version=self._kafka_api_version,
             )
         return self._topic_mng
 
@@ -405,32 +602,46 @@ class Application(Router):
         if replication_factor is not None:
             self._replication_factor = replication_factor
 
+    async def publish_and_wait(
+        self, stream_id: str, data: BaseModel, key: Optional[bytes] = None
+    ) -> aiokafka.structs.ConsumerRecord:
+        return await (await self.publish(stream_id, data, key))
+
     async def publish(
         self, stream_id: str, data: BaseModel, key: Optional[bytes] = None
     ) -> Awaitable[aiokafka.structs.ConsumerRecord]:
-        if not self._intialized:
+        if not self._initialized:
             async with self.get_lock("_"):
                 await self.initialize()
 
         schema_key = getattr(data, "__key__", None)
         if schema_key not in self._schemas:
-            raise UnregisteredSchemaException(model=data)
+            # do not require key
+            schema_key = f"{data.__class__.__name__}:1"
         data_ = data.dict()
 
         topic_id = self.topic_mng.get_topic_id(stream_id)
         async with self.get_lock(stream_id):
             if not await self.topic_mng.topic_exists(topic_id):
                 reg = self.get_schema_reg(data)
+                retention_ms = None
+                if reg is not None and reg.retention is not None:
+                    retention_ms = reg.retention * 1000
                 await self.topic_mng.create_topic(
                     topic_id,
                     replication_factor=self._replication_factor,
-                    retention_ms=reg.retention * 1000 if reg.retention is not None else None,
+                    retention_ms=retention_ms,
                 )
 
-        logger.info(f"Sending kafka msg: {stream_id}")
+        return await self.raw_publish(
+            stream_id, orjson.dumps({"schema": schema_key, "data": data_}), key
+        )
 
+    async def raw_publish(
+        self, stream_id: str, data: bytes, key: Optional[bytes] = None
+    ) -> Awaitable[aiokafka.structs.ConsumerRecord]:
+        logger.debug(f"Sending kafka msg: {stream_id}")
         producer = await self._get_producer()
-
         tracer = opentracing.tracer
         headers: Optional[List[Tuple[str, bytes]]] = None
         if isinstance(tracer.scope_manager, ContextVarsScopeManager):
@@ -444,32 +655,59 @@ class Application(Router):
                 )
                 headers = [(k, v.encode()) for k, v in carrier.items()]
 
-        fut = await producer.send(
-            topic_id,
-            value=orjson.dumps({"schema": schema_key, "data": data_}),
-            key=key,
-            headers=headers,
-        )
+        if not self.producer_healthy():
+            raise ProducerUnhealthyException(self._producer)  # type: ignore
 
-        fut.add_done_callback(partial(_published_callback, topic_id))  # type: ignore
+        topic_id = self.topic_mng.get_topic_id(stream_id)
+        start_time = time.time()
+        with watch_publish(topic_id):
+            fut = await producer.send(topic_id, value=data, key=key, headers=headers,)
+
+        fut.add_done_callback(partial(_published_callback, topic_id, start_time))  # type: ignore
         return fut
 
     async def flush(self) -> None:
         if self._producer is not None:
             await self._producer.flush()
 
-    def get_schema_reg(self, model_or_def: BaseModel) -> SchemaRegistration:
-        key = model_or_def.__key__  # type: ignore
-        return self._schemas[key]
+    def get_schema_reg(self, model_or_def: BaseModel) -> Optional[SchemaRegistration]:
+        try:
+            key = model_or_def.__key__  # type: ignore
+            return self._schemas[key]
+        except (AttributeError, KeyError):
+            return None
+
+    def producer_healthy(self) -> bool:
+        """
+        It's possible for the producer to be unhealthy while we're still sending messages to it.
+        """
+        if self._producer is not None and self._producer._sender.sender_task is not None:
+            return not self._producer._sender.sender_task.done()
+        return True
+
+    def consumer_factory(self, group_id: str) -> aiokafka.AIOKafkaConsumer:
+        return aiokafka.AIOKafkaConsumer(
+            bootstrap_servers=cast(List[str], self._kafka_servers),
+            loop=asyncio.get_event_loop(),
+            group_id=group_id,
+            api_version=self._kafka_api_version,
+            enable_auto_commit=False,
+            **{k: v for k, v in self.kafka_settings.items() if k in _aiokafka_consumer_settings},
+        )
+
+    def producer_factory(self) -> aiokafka.AIOKafkaProducer:
+        return aiokafka.AIOKafkaProducer(
+            bootstrap_servers=cast(List[str], self._kafka_servers),
+            loop=asyncio.get_event_loop(),
+            api_version=self._kafka_api_version,
+            **{k: v for k, v in self.kafka_settings.items() if k in _aiokafka_producer_settings},
+        )
 
     async def _get_producer(self) -> aiokafka.AIOKafkaProducer:
         if self._producer is None:
-            self._producer = aiokafka.AIOKafkaProducer(
-                bootstrap_servers=cast(List[str], self._kafka_servers),
-                loop=asyncio.get_event_loop(),
-                api_version=self._kafka_api_version,
-            )
-            await self._producer.start()
+            self._producer = self.producer_factory()
+            with watch_kafka("producer_start"):
+                await self._producer.start()
         return self._producer
 
     async def initialize(self) -> None:
@@ -488,24 +726,35 @@ class Application(Router):
                             else None,
                         )
 
-        self._intialized = True
+        self._initialized = True
 
     async def finalize(self) -> None:
         await self._call_event_handlers("finalize")
 
         if self._producer is not None:
-            await self._producer.flush()
-            await self._producer.stop()
+            with watch_kafka("producer_flush"):
+                await self._producer.flush()
+            with watch_kafka("producer_stop"):
+                await self._producer.stop()
+
+        if self._subscription_consumers_tasks:
+            for task in self._subscription_consumers_tasks:
+                if not task.done():
+                    task.cancel()
+
+            for task in self._subscription_consumers_tasks:
+                try:
+                    await task
+                except asyncio.CancelledError:
+                    ...
 
         if self._topic_mng is not None:
             await self._topic_mng.finalize()
 
         self._producer = None
-        self._intialized = False
+        self._initialized = False
         self._topic_mng = None
 
-        self._intialized = False
-
     async def __aenter__(self) -> "Application":
         await self.initialize()
         return self
@@ -513,7 +762,7 @@ class Application(Router):
     async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
         await self.finalize()
 
-    async def consume_for(self, num_messages: int, *, seconds: Optional[int] = None) -> None:
+    async def consume_for(self, num_messages: int, *, seconds: Optional[int] = None) -> int:
 
         consumed = 0
 
@@ -546,21 +795,31 @@ class Application(Router):
             for fut in futures:
                 if not fut.done():
                     fut.cancel()
+        return consumed
 
     async def consume_forever(self) -> None:
         self._subscription_consumers = []
+        self._subscription_consumers_tasks = []
 
         for subscription in self._subscriptions:
             consumer = SubscriptionConsumer(self, subscription)
             self._subscription_consumers.append(consumer)
 
         try:
-            futures = [asyncio.create_task(c()) for c in self._subscription_consumers]
-            await asyncio.gather(*futures)
+            self._subscription_consumers_tasks = [
+                asyncio.create_task(c()) for c in self._subscription_consumers
+            ]
+            await asyncio.gather(*self._subscription_consumers_tasks)
         finally:
-            for fut in futures:
-                if not fut.done():
-                    fut.cancel()
+            for task in self._subscription_consumers_tasks:
+                if not task.done():
+                    task.cancel()
+
+            for task in self._subscription_consumers_tasks:
+                try:
+                    await task
+                except asyncio.CancelledError:
+                    ...
 
 
 class CustomConsumerRebalanceListener(aiokafka.ConsumerRebalanceListener):
@@ -584,27 +843,18 @@ class CustomConsumerRebalanceListener(aiokafka.ConsumerRebalanceListener):
         starting_pos = await self.app.topic_mng.list_consumer_group_offsets(
             self.group_id, partitions=assigned
         )
-        logger.info(f"Partitions assigned: {assigned}")
+        logger.debug(f"Partitions assigned: {assigned}")
 
         for tp in assigned:
+            CONSUMER_REBALANCED.labels(
+                stream_id=tp.topic, partition=tp.partition, group_id=self.group_id,
+            ).inc()
             if tp not in starting_pos or starting_pos[tp].offset == -1:
                 # detect if we've never consumed from this topic before
                 # decision right now is to go back to beginning
                 # and it's unclear if this is always right decision
                 await self.consumer.seek_to_beginning(tp)
 
-            # XXX go back one message
-            # unclear if this is what we want to do...
-            # try:
-            #     position = await self.consumer.position(tp)
-            #     offset = position - 1
-            # except aiokafka.errors.IllegalStateError:
-            #     offset = -1
-            # if offset > 0:
-            #     self.consumer.seek(tp, offset)
-            # else:
-            #     await self.consumer.seek_to_beginning(tp)
-
 
 cli_parser = argparse.ArgumentParser(description="Run kafkaesk worker.")
 cli_parser.add_argument("app", help="Application object")
@@ -616,7 +866,7 @@ cli_parser.add_argument("--api-version", help="Kafka API Version")
 
 def _close_app(app: Application, fut: asyncio.Future) -> None:
     if not fut.done():
-        logger.info("Cancelling consumer from signal")
+        logger.debug("Cancelling consumer from signal")
         fut.cancel()
 
 
@@ -627,7 +877,7 @@ async def run_app(app: Application) -> None:
         for signame in {"SIGINT", "SIGTERM"}:
             loop.add_signal_handler(getattr(signal, signame), partial(_close_app, app, fut))
         await fut
-        logger.info("Exiting consumer")
+        logger.debug("Exiting consumer")
 
 
 def run(app: Optional[Application] = None) -> None:
@@ -652,7 +902,7 @@ def run(app: Optional[Application] = None) -> None:
             app.configure(api_version=opts.api_version)
 
     try:
-        logger.info(f"Running kafkaesk consumer {app}")
+        logger.debug(f"Running kafkaesk consumer {app}")
         asyncio.run(run_app(app))
     except asyncio.CancelledError:
-        logger.info("Closing because task was exited")
+        logger.debug("Closing because task was exited")
