@@ -1,6 +1,8 @@
 from pydantic import BaseModel
 from typing import TYPE_CHECKING
 
+import aiokafka
+
 if TYPE_CHECKING:
     from .app import SchemaRegistration
     from .app import SubscriptionConsumer
@@ -47,3 +49,12 @@ class ConsumerUnhealthyException(Exception):
     def __init__(self, subscriber_consumer: SubscriptionConsumer, reason: str):
         self.subscriber_consumer = subscriber_consumer
         self.reason = reason
+
+
+class AutoCommitError(ConsumerUnhealthyException):
+    ...
+
+
+class ProducerUnhealthyException(Exception):
+    def __init__(self, producer: aiokafka.AIOKafkaProducer):
+        self.producer = producer
